package com.analogbit.puzzletools.client.slidingblocks.gui;
import java.util.ArrayList;

import com.allen_sauer.gwt.dnd.client.DragContext;
import com.allen_sauer.gwt.dnd.client.DragEndEvent;
import com.allen_sauer.gwt.dnd.client.DragHandler;
import com.allen_sauer.gwt.dnd.client.DragStartEvent;
import com.allen_sauer.gwt.dnd.client.VetoDragException;
import com.allen_sauer.gwt.dnd.client.drop.GridConstrainedDropController;
import com.analogbit.puzzletools.client.slidingblocks.state.Block;
import com.analogbit.puzzletools.client.slidingblocks.state.BlockShape;
import com.analogbit.puzzletools.client.slidingblocks.state.Grid;
import com.analogbit.puzzletools.client.slidingblocks.state.GridWithListeners;
import com.analogbit.puzzletools.client.util.Color;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.ui.Anchor;
import com.google.gwt.user.client.ui.DecoratedPopupPanel;
import com.google.gwt.user.client.ui.VerticalPanel;

/**
 * The grid display that is used to build the initial grid
 */
public class GridBuilderDisplay extends GridDisplay implements GridDisplayListener {
	private Grid gridBuilding = new GridWithListeners(4, 4);
	private Block blockCreating = null;
	private Color nextColor = null;
	private int lastGridX = -1;
	private int lastGridY = -1;
	private ArrayList<GridBuilderListener> builderListeners = new ArrayList<GridBuilderListener>();
	
	/**
	 * Basic constructor
	 */
	public GridBuilderDisplay() {
		getDragController().registerDropController(new GridConstrainedDropController(this, GridBlock.CELL_SIZE, GridBlock.CELL_SIZE));
		getDragController().addDragHandler(new CellDragHandler());
		getNextColor();
		setGridDisplaying(gridBuilding);
		fireGridChanged();
		addGridListener(this);
	}
	
	/**
	 * Tells the listeners the grid changed.
	 */
	protected void fireGridChanged() {
		for (GridBuilderListener listener : builderListeners) {
			listener.onGridBuildingChanged();
		}
	}
	
	/**
	 * Randomly generates the color of the next block to put down.
	 * Preloads the next color to be returned.
	 * @return New random color.
	 */
	private Color getNextColor() {
		Color result = nextColor;
		nextColor = Color.randomBright();
		GridBlock.preloadColor(nextColor);
		return result;
	}

	@Override
	public void onGridMouseDown(Block blockDown, int gridX, int gridY, int clientX, int clientY) {
		//Window.setTitle("onGridMouseDown: " + blockDown);
		if (blockDown == null) {
			
			blockCreating = new Block(gridX, gridY, new BlockShape(1, 1,
					new boolean[] { true }), getNextColor(), false);
			gridBuilding.addBlock(blockCreating);
			fireGridChanged();
			DOM.setCapture(this.getElement());
		}
	}
	
	@Override
	public void onGridMouseMove(int gridX, int gridY, int clientX, int clientY) { 
		if (lastGridX == gridX && lastGridY == gridY) {
			return;
		}
		
		if (blockCreating != null) {
			int minX = Math.min(gridX, blockCreating.getLeft());
			int minY = Math.min(gridY, blockCreating.getTop());
			int maxX = Math.max(gridX+1, blockCreating.getRight());
			int maxY = Math.max(gridY+1, blockCreating.getBottom());
			
			int width = maxX - minX;
			int height = maxY - minY;
			
			boolean[] bitArray = new boolean[width*height];
			
			int oldOffsetX = blockCreating.getLeft() - minX;
			int oldOffsetY = blockCreating.getTop() - minY;
			boolean[] oldBitArray = blockCreating.getBlockBits();
			for (int y=0; y<blockCreating.getHeight(); y++) {
				for (int x=0; x<blockCreating.getWidth(); x++) {
					bitArray[(x+oldOffsetX) + (y+oldOffsetY)*width] = oldBitArray[x + y*blockCreating.getWidth()];
				}
			}
			bitArray[gridX - minX + (gridY - minY)*width] = true;
			
			Block newBlock = new Block(minX, minY, new BlockShape(width, height, bitArray), blockCreating.getColor(), blockCreating.isImmobile());
			if (gridBuilding.replaceBlockIfLegal(blockCreating, newBlock)) {
				blockCreating = newBlock;
				fireGridChanged();
			}
		}
		
		lastGridX = gridX;
		lastGridY = gridY;
	}

	@Override
	public void onGridContextMenu(Block blockDown, int gridX, int gridY, int clientX, int clientY) {
		showContextMenu(clientX, clientY, gridBuilding.getBlockAt(gridX, gridY));
	}
	
	@Override
	public void onGridMouseUp(boolean clicked, int gridX, int gridY, int clientX, int clientY) {
		//mouseup tells us when we are done creating a block
		if (blockCreating != null) {
			clicked = false; // dont show context menu after creating block
			DOM.releaseCapture(this.getElement());
		}
		blockCreating = null;
		
		
		if (clicked) {
			showContextMenu(clientX, clientY, gridBuilding.getBlockAt(gridX,  gridY));
		}
	}
	
	/**
	 * Displays a context menu where the mouse was clicked.
	 * @param clientX X coord 
	 * @param clientY Y coord
	 * @param blockClicked The block we clicked on, so that we can make it contextual
	 */
	private void showContextMenu(int clientX, int clientY, final Block blockClicked) {
		if (blockClicked != null) {
			final DecoratedPopupPanel popup = new DecoratedPopupPanel(true, true);
			popup.getElement().getStyle().setProperty("zIndex", "5");
			VerticalPanel menu = new VerticalPanel();
			
			Anchor deleteLink = new Anchor("Delete");
			deleteLink.setTitle("Delete this block from the grid.");
			Anchor newColorLink = new Anchor("New Random Color");
			newColorLink.setTitle("Randomly chooses a new color for this block.");
			Anchor lockLink = new Anchor(blockClicked.isImmobile() ? "Unlock" : "Lock in Place");
			lockLink.setTitle("A locked block will not be moved during solving.");
			
			deleteLink.addClickHandler(new ClickHandler() {
				public void onClick(ClickEvent event) {
					gridBuilding.removeBlock(blockClicked);
					fireGridChanged();
					popup.hide();
				}		
			});
			
			newColorLink.addClickHandler(new ClickHandler() {
				public void onClick(ClickEvent event) {
					gridBuilding.setBlockColor(blockClicked, getNextColor());
					fireGridChanged();
					popup.hide();
				}
			});
			
			lockLink.addClickHandler(new ClickHandler() {
				public void onClick(ClickEvent event) {
					gridBuilding.setBlockImmobile(blockClicked, !blockClicked.isImmobile());
					fireGridChanged();
					popup.hide();
				}
			});
			
			menu.add(deleteLink);
			menu.add(newColorLink);
			menu.add(lockLink);
			popup.setWidget(menu);
			popup.setPopupPosition(clientX, clientY);
			popup.show();
		}
	}
	
	/**
	 * Gets the grid that we built.
	 * @return The grid built.
	 */
	public Grid getGrid() {
		return gridBuilding;
	}
	
	/**
	 * All blocks will be dragable during initial setup.
	 */
	@Override
	public boolean isDraggable(GridBlock block) {
		return true;
	}

	public void addGridBuilderListener(GridBuilderListener listener) {
		builderListeners.add(listener);		
	}
	
	public boolean removeGridBuilderListener(GridBuilderListener listener) {
		return builderListeners.remove(listener);
	}

	/**
	 * clears and creates a new grid for building
	 * @param width Width of the new grid
	 * @param height Height of the new grid
	 */
	public void newGrid(int width, int height) {
		gridBuilding = new GridWithListeners(width, height);
		setGridDisplaying(gridBuilding);		
	}

	/**
	 * Handles the drag events for the grid builder. 
	 */
	private class CellDragHandler implements DragHandler {

		public void onDragEnd(DragEndEvent event) {
			DragContext context = event.getContext();
			
			//use this instead of context.desiredDraggable{X,Y} because the dropHandler may have corrected the pos
			int x = (context.draggable.getAbsoluteLeft() - getAbsoluteLeft() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			int y = (context.draggable.getAbsoluteTop() - getAbsoluteTop() + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			
			getGrid().moveBlockIfLegal(((GridBlock)context.draggable).getSourceBlock(), x, y);
			//by default the drop moves the block, lets make sure it put it in the right place
			//corrects for the drop if the move was illegal
			updateGridBlockPosition((GridBlock) context.draggable);
		}

		public void onDragStart(DragStartEvent event) {
		}

		public void onPreviewDragEnd(DragEndEvent event)
				throws VetoDragException {
		}

		public void onPreviewDragStart(DragStartEvent event)
				throws VetoDragException {
			DragContext context = event.getContext();
			//this information is stale (gwt-dnd issue #44) 
			//int cellX = (context.desiredDraggableX/*-getAbsoluteLeft()*/ + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			//int cellY = (context.desiredDraggableY/*-getAbsoluteTop()*/ + GridBlock.CELL_SIZE/2) / GridBlock.CELL_SIZE;
			
			int cellX = getCellMouseDownX();
			int cellY = getCellMouseDownY();

			GridBlock gb = (GridBlock)context.draggable;
			if (!gb.getSourceBlock().hasCellAt(cellX, cellY)) {
				throw new VetoDragException();
			}
		}

	}
	
}
